// Vgm2Midi - Ben Ryves 2006
using System;
using System.Collections.Generic;
using System.Text;

namespace Vgm2Midi {
    partial class Program {

        // This contains the VGM 'player' and MIDI conversion code.
        
        static int VgmDelay = 0; // Delay between commands (in VGM file).

        static bool StillWorking = true; // Flag to say we are still decoding the file

        static int SamplesPassed = 0; // How many 44100Hz samples have been played
        static int LastMidiWrite = 0; // Last time (in samples) MIDI data was written.

        static int VgmDataPointer = 0;

        /// <summary>
        /// Get a byte from the VGM file
        /// </summary>
        static byte GetInstruction() {
            if (VgmDataPointer + 1 >= ToConvert.VgmData.Length) {
                StillWorking = false;
                return 0x00;
            }
            return ToConvert.VgmData[VgmDataPointer++];
        }

        /// <summary>
        /// Grab data from the VGM file and do something with it (eg, write to PSG).
        /// </summary>
        public static void AdvanceThroughVgmFile() {
            SamplesPassed += VgmDelay;
            VgmDelay = 0;
            while (VgmDelay == 0) {
                if (!StillWorking) return;
                byte Instruction = GetInstruction();
                switch (Instruction) {
                    case 0x4F: // GG stereo
                        SN76489.WriteStereoByte(GetInstruction());
                        break;
                    case 0x50: // It's a PSG write
                        SN76489.WriteByte(GetInstruction());
                        break;
                    case 0x51: // YM2413 - reg # = #.
                    case 0x52: // YM2612 - reg # = # on port 0.
                    case 0x53: // YM2612 - reg # = # on port 1.
                    case 0x54: // YM2151 - reg # = #.
                        byte Register = GetInstruction();
                        byte Value = GetInstruction();
                        switch (Instruction) {
                            case 0x51:
                                YM2413.Write(Register, Value);
                                break;
                        }
                        break;
                    case 0x61: // Wait ##
                        VgmDelay = GetInstruction();
                        VgmDelay += GetInstruction() << 8;
                        break;
                    case 0x62: // Wait 735
                        VgmDelay = 735;
                        break;
                    case 0x63: // Wait 882
                        VgmDelay = 882;
                        break;
                    case 0x66: // End of data
                        SN76489.Reset();
                        if (ToConvert.LoopSamples == 0) {
                            StillWorking = false;
                        } else {
                            if (LoopCount < 1 && (!FadingOut || FadeoutTime <= 0.0d)) {
                                StillWorking = false;

                            } else {
                                --LoopCount;
                                VgmDataPointer = 0;
                                int OldSamplesPast = SamplesPassed;
                                while (VgmDataPointer < ToConvert.LoopOffset - 0x1C) {
                                    AdvanceThroughVgmFile();
                                }
                                SamplesPassed = OldSamplesPast;
                                if (LoopCount == 0 && FadeoutTime != 0.0d) FadingOut = true;
                            }

                        }
                        break;
                    case 0x67: // Data block
                        // ?
                        break;
                    case 0x70:
                    #region snip!
                    case 0x71:
                    case 0x72:
                    case 0x73:
                    case 0x74:
                    case 0x75:
                    case 0x76:
                    case 0x77:
                    case 0x78:
                    case 0x79:
                    case 0x7A:
                    case 0x7B:
                    case 0x7C:
                    case 0x7D:
                    case 0x7E:
                    #endregion
                    case 0x7F:
                        VgmDelay = (Instruction & 0xF) + 1;
                        break;
                    case 0x80:
                    #region snip!
                    case 0x81:
                    case 0x82:
                    case 0x83:
                    case 0x84:
                    case 0x85:
                    case 0x86:
                    case 0x87:
                    case 0x88:
                    case 0x89:
                    case 0x8A:
                    case 0x8B:
                    case 0x8C:
                    case 0x8D:
                    case 0x8E:
                    #endregion
                    case 0x8F:
                        // Fetch byte from data block, write to address $2A on port 0 of YM2612
                        // ... then delay
                        VgmDelay = Instruction & 0xF;
                        break;
                    case 0xE0:
                        uint DataBlockPointer = 0;
                        for (int i = 0; i < 4; i++) {
                            DataBlockPointer >>= 8;
                            DataBlockPointer |= (uint)(GetInstruction() << 24);
                        }
                        break;
                    default:
                        // Reserved:
                        if (Instruction >= 0x30 && Instruction <= 0x4E) {
                            GetInstruction();
                        } else if (Instruction >= 0x55 && Instruction <= 0x5F) {
                            GetInstruction();
                            GetInstruction();
                        } else if (Instruction >= 0xA0 && Instruction <= 0xBF) {
                            GetInstruction();
                            GetInstruction();
                        } else if (Instruction >= 0xC0 && Instruction <= 0xDF) {
                            GetInstruction();
                            GetInstruction();
                            GetInstruction();
                        } else if (Instruction >= 0xE1 && Instruction <= 0xFF) {
                            GetInstruction();
                            GetInstruction();
                            GetInstruction();
                            GetInstruction();
                        }
                        break;
                }
            }
        }


        public enum MidiMessage : byte {
            KeyRelease = 0x80,
            KeyPress = 0x90,
            KeyAfterTouch = 0xA0,
            ControlChange = 0xB0,
            PatchChange = 0xC0,
            ChannelAfterTouch = 0xD0,
            PitchWheelChange = 0xE0
        };
        private static void KeyPress(int Channel, int Key, int Velocity) {
            WriteTime();
            MidiData.Add((byte)((int)MidiMessage.KeyPress | Channel));
            MidiData.Add((byte)(Key & 0x7F));
            MidiData.Add((byte)(Velocity & 0x7F));
        }
        private static void KeyRelease(int Channel, int Key) {
            WriteTime();
            MidiData.Add((byte)((int)MidiMessage.KeyRelease | Channel));
            MidiData.Add((byte)(Key & 0x7F));
            MidiData.Add((byte)64);
        }
        private static void PitchWheelChange(int Channel, int NewValue) {
            WriteTime();
            MidiData.Add((byte)((int)MidiMessage.PitchWheelChange | Channel));
            MidiData.Add((byte)(NewValue & 0x7F));
            MidiData.Add((byte)((NewValue >> 7) & 0x7F));
        }
        private static void PatchChange(int Channel, int Patch) {
            WriteTime();
            MidiData.Add((byte)((int)MidiMessage.PatchChange | Channel));
            MidiData.Add((byte)(Patch & 0x7F));
        }
        public static void Retouch(int Channel, int Key, int Velocity) {
            WriteTime();
            MidiData.Add((byte)((int)MidiMessage.KeyAfterTouch | Channel));
            MidiData.Add((byte)(Key & 0x7F));
            MidiData.Add((byte)(Velocity & 0x7F));
        }
        public static void Pan(int Channel, int Pan) {
            WriteTime();
            MidiData.Add((byte)((int)MidiMessage.ControlChange | Channel));
            MidiData.Add(10);
            MidiData.Add((byte)Pan);  
        }
        public static void SetVolume(int Channel, int Volume) {
            WriteTime();
            MidiData.Add((byte)((int)MidiMessage.ControlChange | Channel));
            MidiData.Add(7);
            MidiData.Add((byte)Volume);
        }
        private static void WriteTime() {
            MidiTimeDelay = SamplesPassed - LastMidiWrite;
            WriteVariableLengthValue((uint)MidiTimeDelay);
            
            LastMidiWrite = SamplesPassed;
        }

        static void WriteVariableLengthValue(uint value) {
            uint buffer = value & 0x7F;

            while ((value >>= 7) != 0) {
                buffer <<= 8;
                buffer |= ((value & 0x7F) | 0x80);
            }

            for (; ; ) {
                MidiData.Add((byte)(buffer & 0xFF));
                if ((buffer & 0x80) != 0) {
                    buffer >>= 8;
                } else {
                    break;
                }
            }
        }

        private static int LoopCount = 1;
        private static double FadeoutTime = 0.0d;
        private static double OriginalFadeoutTime = 0.0d;
        private static bool FadingOut = false;

        static int[] LastKeysPressed = new int[] { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
        static int[] LastLevelPlayed = new int[] { 0xF, 0xF, 0xF, 0xF };
        static int[] LastChannelNumber = new int[] { 0, 1, 2, 3 };
        static int[] LastPitchWheel = new int[] { 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000 };
        static int[] LastPan = new int[] { 64, 64, 64, 64 };
        static int[] LastVolume = new int[] { 127, 127, 127, 127 };
        static int Instrument0 = 80;
        static int Instrument1 = 80;
        static int Instrument2 = 80;
        static int Instrument3 = 80;


        static bool[] WasKeyOn = new bool[] { false, false, false, false, false, false, false, false, false };

        static int HighPercussion = 48;
        static int MidPercussion = 43;
        static int LowPercussion = 35;


        static int[] Volumes = new int[] { 127, 127, 127, 127 };

        static int[] FakeStereo = new int[] { 0, 32, -32, 0 };

        static int[] LastInstrument = new int[] { -1, -1, -1, -1, -1, -1, -1, -1, -1 };
        /*  0	Original	
            1	Violin	
            2	Guitar	
            3	Piano	
            4	Flute	
            5	Clarinet	
            6	Oboe	
            7	Trumpet	
            8	Organ	
            9	Horn	
            10	Synthesizer	
            11	Harpsichord	
            12	Vibraphone	
            13	Synthesizer Bass	
            14	Acoustic Bass	
            15	Electric Guitar
        */
        static int[] InstrumentMap = new int[] { 4, 40, 24, 0, 73, 71, 68, 56, 19, 60, 81, 6, 11, 38, 32, 27 };
        static int[] LastOpllVolume = new int[] { 0, 0, 0, 0, 0, 0, 0, 0, 0 };

        static int A440Key = 69;
        static decimal FreqScaleFactor = 1;
        static int MidiTimeDelay = 0;

        /// <summary>
        /// Write any changes (eg new notes) to the MIDI file
        /// </summary>
        static void UpdateMidi() {
            if (FadingOut) {
                FadeoutTime -= ((double)(SamplesPassed - LastMidiWrite) / 44.1d);
                if (FadeoutTime < 0) {
                    FadeoutTime = 0;
                    StillWorking = false;
                }
            }

            if (ToConvert.PsgSpeed != 0) {
                // SN76489 (PSG) fun
                for (int i = 0; i < 4; i++) {
                    if (Volumes[i] == 0) continue;
                    if (SN76489.ToneReg[i == 3 ? 2 : i] != 0) {

                        decimal Frequency = ((decimal)ToConvert.PsgSpeed * FreqScaleFactor) / (32.0m * (decimal)SN76489.ToneReg[i == 3 ? 2 : i]);
                        if (i == 3) {
                            Frequency /= ToConvert.LsfrWidth;
                        }
                        decimal Key = 12.0m * (decimal)Math.Log((double)Frequency, 2.0d);

                        int MidiKey = 0;
                        int NewPitchWheel = 0x2000;
                        int ChannelNumber = i;

                        if (i == 3 && ((SN76489.ToneReg[3] & 0x3) != 3)) {
                            // Percussion!
                            switch (SN76489.ToneReg[3] & 0x3) {
                                case 0x00:
                                    MidiKey = HighPercussion;
                                    break;
                                case 0x01:
                                    MidiKey = MidPercussion;
                                    break;
                                case 0x02:
                                    MidiKey = LowPercussion;
                                    break;
                            }
                            ChannelNumber = 9;
                        } else {
                            MidiKey = Math.Max(Math.Min(127, (int)(Math.Round(Key, 0))), 0);
                            decimal Remainder = (Key - (decimal)MidiKey);
                            NewPitchWheel = 0x2000 + (int)(Remainder * 4096.0m);
                        }

                        int Velocity = Math.Min(127, (0xF - SN76489.Levels[i]) * 8);

                        if (SN76489.Levels[i] == 0xF) {
                            // Absolute silence (shhh)
                            if (LastLevelPlayed[i] != 0xF) KeyRelease(i, LastKeysPressed[i]); // If it was 0xF last time, it would have already been released.
                            LastLevelPlayed[i] = SN76489.Levels[i];
                        } else if ((SN76489.Levels[i] < LastLevelPlayed[i]) || (LastKeysPressed[i] != MidiKey)) {
                            // New note being played

                            if ((SN76489.Levels[i] < LastLevelPlayed[i]) && (LastKeysPressed[i] == MidiKey) && (LastLevelPlayed[i] - SN76489.Levels[i] < 4)) {
                                Retouch(ChannelNumber, MidiKey, Velocity);
                            } else {

                                if (LastLevelPlayed[i] != 0xF) KeyRelease(i, LastKeysPressed[i]);

                                KeyPress(ChannelNumber, MidiKey, Velocity);
                            }
                            LastKeysPressed[i] = MidiKey;
                            LastLevelPlayed[i] = SN76489.Levels[i];


                            if (NewPitchWheel != LastPitchWheel[i]) {
                                PitchWheelChange(ChannelNumber, NewPitchWheel);
                                LastPitchWheel[i] = NewPitchWheel;
                            }

                        } else {
                            // Note is fading away:
                            //Retouch(ChannelNumber, MidiKey, Velocity);
                            if (NewPitchWheel != LastPitchWheel[i]) {
                                PitchWheelChange(ChannelNumber, NewPitchWheel);
                                LastPitchWheel[i] = NewPitchWheel;
                            }

                        }
                        LastLevelPlayed[i] = SN76489.Levels[i];


                        LastChannelNumber[i] = ChannelNumber;
                        int StereoStatus = SN76489.Stereo[i, 1] + SN76489.Stereo[i, 0] * 2;
                        int PanAmount = 64;
                        int Volume = 127;
                        switch (StereoStatus) {
                            case 0:
                                Volume = 0;
                                break;
                            case 1:
                                PanAmount = 0;
                                break;
                            case 2:
                                PanAmount = 127;
                                break;
                            case 3:
                                break;
                        }
                        PanAmount += FakeStereo[i];
                        PanAmount = Math.Max(Math.Min(PanAmount, 127), 0);
                        if (LastPan[i] != PanAmount) {
                            Pan(i, PanAmount);
                            if (i == 3) Pan(9, PanAmount);
                            LastPan[i] = PanAmount;
                        }

                        if (FadeoutTime != OriginalFadeoutTime) {
                            // We're fading out
                            Volume = (int)((((double)Volume) * FadeoutTime) / OriginalFadeoutTime);
                        }

                        Volume = (int)Math.Max(0, Math.Min(127, Math.Round((double)(Volumes[i] * Volume) / 127.0d, 0)));

                        if (LastVolume[i] != Volume) {
                            SetVolume(i, Volume);
                            if (i == 3) SetVolume(9, Volume);
                            LastVolume[i] = Volume;
                        }

                    }
                }
            }



            if (ToConvert.Y2413Speed != 0) {
                // YM2413 (OPLL) fun
                int MidiChannelNumber = 4; // PSG takes up 0-3

                int NumberOfToneChannels = 6;// YM2413.RhythmMode ? 6 : 9;
                for (int i = 0; i < NumberOfToneChannels; ++i) {
                    YM2413.OpllChannel c = YM2413.OpllChannels[i];

                    decimal Frequency = ((decimal)(c.PhaseIncrement[1]) * FreqScaleFactor * 44100.0m);
                    int MidiKey = 0;
                    int NewPitchWheel = 0x2000;
                    int NewVolume = Math.Min(127, c.Volume * 32);
                    if (Frequency != 0.0m) {
                        decimal Key = 12.0m * (decimal)Math.Log((double)Frequency, 2.0d);
                        MidiKey = (int)Math.Round(Key, 0);
                        decimal Remainder = (Key - (decimal)MidiKey);
                        NewPitchWheel = 0x2000 + (int)(Remainder * 4096.0m);
                    } else {
                        continue;
                    }


                    if ((!WasKeyOn[i] && c.KeyOn)) { //|| (c.KeyOn && (LastKeysPressed[i + 4] != MidiKey))) {
                        // Key is being pressed
                        if (LastInstrument[i] != c.InstrumentIndex) {
                            LastInstrument[i] = c.InstrumentIndex;
                            PatchChange(MidiChannelNumber, InstrumentMap[i]);
                        }

                        KeyPress(MidiChannelNumber, MidiKey, NewVolume);
                        
                        if (LastPitchWheel[i + 4] != NewPitchWheel) {
                            PitchWheelChange(MidiChannelNumber, NewPitchWheel);
                            LastPitchWheel[i + 4] = NewPitchWheel;
                        }
                        LastKeysPressed[i + 4] = MidiKey;
                        WasKeyOn[i] = true;

                    } else if (WasKeyOn[i] && !c.KeyOn) {
                        // Key is being released
                        KeyRelease(MidiChannelNumber, LastKeysPressed[i + 4]);
                        WasKeyOn[i] = false;
                    } 
                    MidiChannelNumber++;
                    if (MidiChannelNumber == 9) MidiChannelNumber = 10; // Skip over the percussion channel
                }
                if (YM2413.RhythmMode) {
                    // DRUMS!
                    
                }

            }

        }

        


        
    }
}
