// Vgm2Midi - Ben Ryves 2006
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
namespace Vgm2Midi {
    partial class Program {
        
        static VgmFile ToConvert = null; // VGM file to convert
        static BinaryWriter OutMidiFile = null; // MIDI file writer
        static List<byte> MidiData; // MIDI file track data

        static void Main(string[] args) {
            Console.WriteLine("Vgm2Midi - Ben Ryves 2006");
            Console.WriteLine("-------------------------");
            if (File.Exists("defaults.txt")) {
                try {
                    string[] Lines;
                    using (TextReader T = new StreamReader("defaults.txt")) {
                        Lines = T.ReadToEnd().Split('\n');
                    }
                    foreach (string L in Lines) {
                        string[] Parts = L.Split(';');
                        string ToUse = Parts[0].Trim().ToLower();
                        if (ToUse != "") {
                            string[] VarSplit = ToUse.Split('=');
                            if (VarSplit.Length == 2) {
                                string VarToSet = VarSplit[0].Trim();
                                int ValueToSet = Convert.ToInt32(VarSplit[1].Trim());
                                switch (VarToSet) {
                                    case "instrument0": Instrument0 = (ValueToSet-1) & 0x7F; break;
                                    case "instrument1": Instrument1 = (ValueToSet - 1) & 0x7F; break;
                                    case "instrument2": Instrument2 = (ValueToSet - 1) & 0x7F; break;
                                    case "instrument3": Instrument3 = (ValueToSet - 1) & 0x7F; break;
                                    case "percussion0": HighPercussion = ValueToSet & 0x7F; break;
                                    case "percussion1": MidPercussion = ValueToSet & 0x7F; break;
                                    case "percussion2": LowPercussion = ValueToSet & 0x7F; break;
                                    case "a440": A440Key = ValueToSet & 0x7F; break;
                                    case "stereo0": FakeStereo[0] = ValueToSet; break;
                                    case "stereo1": FakeStereo[1] = ValueToSet; break;
                                    case "stereo2": FakeStereo[2] = ValueToSet; break;
                                    case "stereo3": FakeStereo[3] = ValueToSet; break;
                                    case "loop": LoopCount = ValueToSet; break;
                                    case "fadeout": FadeoutTime = (double)ValueToSet; break;
                                    case "volume0": Volumes[0] = ValueToSet; break;
                                    case "volume1": Volumes[1] = ValueToSet; break;
                                    case "volume2": Volumes[2] = ValueToSet; break;
                                    case "volume3": Volumes[3] = ValueToSet; break;
                                    default:
                                        Console.WriteLine("Variable " + VarToSet + " not understood.");
                                        break;
                                    }
                            }
                        }
                        
                    }
                    Console.WriteLine("Defaults loaded.");
                } catch (Exception ex) {
                    Console.WriteLine("Error in defaults.txt: " + ex.Message);
                }
            }
            if (args.Length == 0) {
                Console.WriteLine("Please see readme.htm for instructions!");
                return;
            }
            Queue<string> CommandLineArgs = new Queue<string>();
            foreach (string s in args) {
                CommandLineArgs.Enqueue(s);                
            }

            string ToConvertFilename = "";
            try {
                while (CommandLineArgs.Count > 0) {
                    string Command = CommandLineArgs.Dequeue();
                    switch (Command.ToLower()) {
                        case "-i":
                            if (CommandLineArgs.Count < 4) {
                                Console.WriteLine("-i switch requires four following arguments!");
                                return;
                            }
                            Instrument0 = (Convert.ToInt32(CommandLineArgs.Dequeue()) - 1) & 0x7F;
                            Instrument1 = (Convert.ToInt32(CommandLineArgs.Dequeue()) - 1) & 0x7F;
                            Instrument2 = (Convert.ToInt32(CommandLineArgs.Dequeue()) - 1) & 0x7F;
                            Instrument3 = (Convert.ToInt32(CommandLineArgs.Dequeue()) - 1) & 0x7F;
                            break;
                        case "-p":
                            if (CommandLineArgs.Count < 3) {
                                Console.WriteLine("-p switch requires 3 following arguments!");
                                return;
                            }
                            HighPercussion = Convert.ToInt32(CommandLineArgs.Dequeue()) & 0x7F;
                            MidPercussion = Convert.ToInt32(CommandLineArgs.Dequeue()) & 0x7F;
                            LowPercussion = Convert.ToInt32(CommandLineArgs.Dequeue()) & 0x7F;
                            break;
                        case "-a":
                            if (CommandLineArgs.Count < 1) {
                                Console.WriteLine("-a switch requires a following argument!");
                                return;
                            }
                            A440Key = Convert.ToInt32(CommandLineArgs.Dequeue()) & 0x7F;
                            break;
                        case "-s":
                            if (CommandLineArgs.Count < 4) {
                                Console.WriteLine("-s switch requires four following arguments!");
                                return;
                            }
                            for (int i = 0; i < 4; i++) FakeStereo[i] = Convert.ToInt32(CommandLineArgs.Dequeue());
                            break;
                        case "-v":
                            if (CommandLineArgs.Count < 4) {
                                Console.WriteLine("-v switch requires four following arguments!");
                                return;
                            }
                            for (int i = 0; i < 4; i++) Volumes[i] = Convert.ToInt32(CommandLineArgs.Dequeue());
                            break;
                        case "-l":
                            if (CommandLineArgs.Count < 1) {
                                Console.WriteLine("-l switch requires a following argument!");
                                return;
                            }
                            LoopCount = Convert.ToInt32(CommandLineArgs.Dequeue());
                            break;
                        case "-f":
                            if (CommandLineArgs.Count < 1) {
                                Console.WriteLine("-f switch requires a following argument!");
                                return;
                            }
                            LoopCount = Convert.ToInt32(CommandLineArgs.Dequeue());
                            break;
                        default:
                            ToConvertFilename = Command.Trim('"');
                            break;
                    }
                }
            } catch (Exception ex) {
                Console.WriteLine("Command-line error: " + ex.Message);
                return;
            }

            if (ToConvertFilename == "") {
                Console.WriteLine("No filename specified!");
                return;
            }


            FreqScaleFactor = (decimal)Math.Pow(2.0d, (double)((decimal)A440Key / 12.0m)) / 440.0m;
            OriginalFadeoutTime = FadeoutTime;

            try {
                ToConvert = new VgmFile(ToConvertFilename);
            } catch (Exception ex) {
                Console.WriteLine("Error loading VGM file: " + ex.Message);
                return;
            }

            Console.WriteLine("Converting: {0} - {1} [{2}]", ToConvert.Game, ToConvert.Track, ToConvert.Author);
            

            string MidiName = Path.GetFileNameWithoutExtension(ToConvertFilename) + ".mid";

            try {
                if (File.Exists(MidiName)) File.Delete(MidiName);
                OutMidiFile = new BinaryWriter(new FileStream(MidiName, FileMode.OpenOrCreate));
            } catch (Exception ex) {
                Console.WriteLine("Could not write output file: " + ex.Message);
                return;
            }

            VgmDataPointer = 0;
            SamplesPassed = 0;
            StillWorking = true;
            MidiData = new List<byte>();

            YM2413.OpllClock = ToConvert.Y2413Speed;

            OutMidiFile.Write("MThd".ToCharArray());
            OutMidiFile.Write(new byte[] { 0x00, 0x00, 0x00, 0x06 }); // Header size (always 6).

            OutMidiFile.Write(new byte[] { 0x00, 0x00 }); // MIDI type (0).
            OutMidiFile.Write(new byte[] { 0x00, 0x01 }); // Number of tracks (always 1 for type 0).
            OutMidiFile.Write(new byte[] { 0x56, 0x22 }); // 22050 resolution

            // Create the track data
            PatchChange(0, Instrument0);
            PatchChange(1, Instrument1);
            PatchChange(2, Instrument2);
            PatchChange(3, Instrument3);

            for (int i = 0; i < 16; i++) {
                SetVolume(i, 127);
                
            }
            
            try {
                SN76489.Reset();
                YM2413.Reset();
                while (StillWorking) {
                    AdvanceThroughVgmFile();
                    UpdateMidi();
                }
            } catch (Exception ex) {
                Console.WriteLine("Conversion failed: " + ex.Message);
            }
            
            // Finish with the end of track message
            WriteTime();
            MidiData.Add(0xFF);
            MidiData.Add(0x2F);
            MidiData.Add(0x00);            

            OutMidiFile.Write("MTrk".ToCharArray()); 
            uint TrackLength = (uint)MidiData.Count;
            for (int i = 0; i < 4; i++) {
                OutMidiFile.Write((byte)(TrackLength >> 24));
                TrackLength <<= 8;
            }
            OutMidiFile.Write(MidiData.ToArray());            
            OutMidiFile.Close();
            Console.WriteLine("Done!");
            
        }
    }
}

