// Vgm2Midi - Ben Ryves 2006
using System;
using System.Collections.Generic;
using System.Text;

namespace Vgm2Midi {
    static class SN76489 {

        // Basic, crippled version of the PSG class I normally use.
        // Doesn't support generating square wave output, but has the basic skeleton functionality I need.

        // Current state
        public static int[] Status = { 0, 0, 0, 0 };		    // Voltage level (0=low, 1=high)
        public static int[,] Stereo = new int[4, 2];         // 1 or 0 (for stereo writes)
        public static int[] Levels = { 15, 15, 15, 15 };	    // Volume level of the 4 channels.
        
        public static int[] ToneReg = { 0, 0, 0, 0 };		// 10-bit tone registers.

        // Internal variables
        private static int[] ToneCounters = { 0, 0, 0, 0 };
        private static int LatchedChannel = 0;
        private static bool LatchedTone = false;
        public static bool NoiseLatch = false;
        private static int NoiseValue = 0;
        private static int NoiseLength = 16;
        public static int NoiseTapped = 0x9;

        // Write to the Game Gear's stereo thingummy
        public static void WriteStereoByte(byte Command) {
            for (int c = 0; c < 2; ++c) {
                for (int r = 0; r < 4; ++r) {
                    Stereo[r, c] = (Command & 1);
                    Command >>= 1;
                }
            }
        }


        // Call to write a byte to the PSG
        public static void WriteByte(byte Command) {
            if ((Command & 0x80) != 0) {
                // It is a LATCH command
                LatchedChannel = (Command >> 5) & 0x3;
                if ((Command & 0x10) == 0) {
                    LatchedTone = true;
                    ToneReg[LatchedChannel] &= 0xFFF0;
                    ToneReg[LatchedChannel] |= (Command & 0xF);
                    // Reset the LSFR?
                    if (LatchedChannel == 3) {
                        NoiseValue = 1 << (NoiseLength - 1);
                    }
                } else {
                    LatchedTone = false;
                    Levels[LatchedChannel] = Command & 0xF;
                }
            } else {
                // It is a DATA command
                if (LatchedTone) {
                    ToneReg[LatchedChannel] &= 0xF;
                    ToneReg[LatchedChannel] |= (Command & 0x3F) << 4;
                } else {
                    Levels[LatchedChannel] = Command & 0xF;
                }
            }
        }

        // Call to reset the PSG's state
        public static void Reset() {
            for (int i = 0; i < 4; ++i) {
                ToneReg[i] = 0;
                Levels[i] = 15;
                Status[i] = 0;
                ToneCounters[i] = 0;
                for (int c = 0; c < 2; ++c) {
                    Stereo[i, c] = 1;
                }
            }
            LatchedChannel = 0;
            LatchedTone = false;
            NoiseLatch = false;
            NoiseValue = 1 << (NoiseLength - 1);
        }

    }
}
